# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/utils.ipynb (unless otherwise specified).

__all__ = ['deep_copy_lmdb']

# Cell
from ifcimglib import imglmdb
import lmdb
import os

# Cell
def deep_copy_lmdb(old_lmdb_env, new_name):
    fd = os.open(new_name, os.O_RDWR)
    old_lmdb_env.copyfd(fd)
    os.close(fd)